
#include "qobjects.h"

#include <cstdlib>
#include <ctime>

int main(int argc, char* argv[]){ 
   
//---------------------------------------initialize parameters

   const int seed = 234562;               // random seed  23456

   const double hbar = 1,                // fundamental constants
                e  = 2.7182818,
                pi = 3.1415927; 
		
   const std::complex<double> ei(0,1); 
   
   const unsigned int N = 15,             // number of fock states 
                      ns = 10000000,       // number of time steps 
                      ns_noise = ns,   
                      nt = 1;             // number of trajectories
   const int          n_out = 500;
   
   const double T = 400,                   // total time 
                dt = T/ns,

                omega = 100,             // *** not currently used
                k = 4,
                coupling = 3,            // interaction strength
                g = k/10000,             // thermal damping rate divided by 4
                xi = 10,                 // temperature factor
                gamma = k/40,            // cpb decay rate
				xic = xi/50;             // temperature factor for CPB


//--------------------------------initialise files

   std::ofstream series("emeas.dat",std::ios_base::trunc);
      
//--------------------------------initialize operators

   qoperator I("identity",N),  sI("identity",2),
             a("destroy",N),   sa("destroy",2),   a_1 = tensor(a,sI), a_2 = tensor(I,sa),
             adag("create",N), sadag("create",2), adag_1 = tensor(adag,sI), adag_2 = tensor(I,sadag),
             E_1 = adag_1*a_1, E_2 = adag_2*a_2, E = adag*a, x = a + adag,
			 I_tot = tensor(I,sI), z_2 = 2*E_2 - I_tot,  
             x_1 = a_1 + adag_1, x_2 = a_2 + adag_2, p_1 = ei*(adag_1 - a_1), 
             x2_1 = x_1*x_1, x2_2 = x_2*x_2;
             
// Hamiltonian    
   qoperator iHdt = (ei*coupling*dt)*(a_1*adag_2 + adag_1*a_2);
   //qoperator iHdt = (ei*omega*dt)*adag_1*a_1 
   //               + (ei*omega*dt)*adag_2*a_2 + (ei*coupling*dt)*x_1*x_2;
   qoperator iH2dt2 = 0.5*iHdt*iHdt;

// observable to measure
   qoperator M_op,M2_op;
   M_op = E_2;
   M2_op = M_op*M_op;
   
// noise operators
   qoperator B_op = x_2;
   qoperator B2_op = B_op*B_op;
   
// other operators   
   qoperator O_op = E_1, O2_op = O_op*O_op;
                
//--------------------------------initial state

   //std::vector< std::complex<double> > psi_vec(N); psi_vec[2] = 1;             
   //qstate psi_1(psi_vec);
   qstate psi_1("coherent",std::sqrt(3.0),N);
   qstate psi_2("up",2);
   qstate psi = tensor(psi_1,psi_2);      
   qmatrix rho(psi);
      
//---------------------------------parameters for Wiener noise (dW's)

   std::vector<double> noise(2);     
   srand(seed); 
   double r8kdt = std::sqrt(8*k*dt);

//------------------------------------------------------------
//                        Simulation
//------------------------------------------------------------

   double EM_op = trace(M_op*rho).real(),  // initialise variables 
          EO_op = trace(O_op*rho).real(),
          Ex    = trace(B_op*rho).real(), 
          EO2   = trace(O2_op*rho).real(),
          EM2   = trace(M2_op*rho).real(),
          VO    = EO2 - EO_op*EO_op,
          VM    = EM2 - EM_op*EM_op,
          d_rec = 0; 
          
   time_t time0, *time_ptr = 0;       // timing initialisation
   time0 = time(time_ptr);
   
   // do multiple trajectories
   for (unsigned int j=0;j<nt;j++) {               
   
     // initial output 
     series << 0 << " " << Ex << " " << EO_op << " " << EM_op << " " 
			<< VO << " " << VM << " " << EM2 << std::endl; 
      
     // integrate a single trajectory
     for (unsigned int i=0;i<ns;i++) {            
          
       if (std::div(i,10000).rem==0) {std::cout << "T = " << i*dt << std::endl;}  // report progress
   
       // Hamiltonian Evolution (2nd order in dt)
       rho +=  iHdt*rho - rho*iHdt - 2*(iHdt*rho*iHdt) + rho*iH2dt2 + iH2dt2*rho; 
       
       make_noise(noise); 
       for (int ni=0;ni<noise.size();ni++) { noise[ni] /= r8kdt; }  
       
       // Measurement Evolution (half order in dt)
       EM_op = trace(M_op*rho).real();       
       d_rec = 4*k*dt*(EM_op + noise[0]);
       rho  += (-k*dt)*(M2_op*rho + rho*M2_op - (2.0*M_op)*rho*M_op) 
             + d_rec*(M_op*rho + rho*M_op);


       // QO Thermal Noise
       rho += ((xi+1)*g*dt)*((2*a_1)*rho*adag_1 - adag_1*a_1*rho - rho*adag_1*a_1) 
              + ((xi)*g*dt)*((2*adag_1)*rho*a_1 - a_1*adag_1*rho - rho*a_1*adag_1);
              
       // BM Thermal Noise
       //rho += (-0.25*ei*g*dt)*( x_1*(p_1*rho + rho*p_1) - (p_1*rho + rho*p_1)*x_1 ) 
       //     - (0.25*g*xi*dt)*(x2_1*rho + rho*x2_1 - 2*x_1*rho*x_1);
	   
	   // CPB decay 
	   rho += ((xic+1)*gamma*dt)*((2*a_2)*rho*adag_2 - adag_2*a_2*rho - rho*adag_2*a_2) 
		        + (xic*gamma*dt)*((2*adag_2)*rho*a_2 - a_2*adag_2*rho - rho*a_2*adag_2);
				
	   // CPB dephasing
	   rho += (2*gamma*dt)*(z_2*rho*z_2 - rho);

       rho.normalise();     
       
       // some expectation values
       EO_op = trace(O_op*rho).real();
       EO2 = trace(O2_op*rho).real();
       EM2 = trace(M2_op*rho).real();
       VO = EO2 - EO_op*EO_op;
       VM = EM2 - EM_op*EM_op;
       
       // output time series
       if (std::div(i,n_out).rem==0) {
         series << i*dt << " " << Ex << " " <<  EO_op << " " << EM_op << " " 
				<< VO << " " << VM << " " << EM2 << std::endl; 
       }


     }
     
   }
         
   //-----------------------
   // output simulation time
   //-----------------------

   std::cout << std::endl << "Simulation took " << difftime(time(time_ptr),time0) 
                          << " seconds" << std::endl;
                          
   series.close();
   	                 
   return 0;
}
